classnames = [
    "airplane",
    "airport",
    "baseball diamond",
    "basketball court",
    "beach",
    "bridge",
    "chaparral",
    "church",
    "circular farmland",
    "cloud",
    "commercial area",
    "dense residential",
    "desert",
    "forest",
    "freeway",
    "golf course",
    "ground track field",
    "harbor",
    "industrial area",
    "intersection",
    "island",
    "lake",
    "meadow",
    "medium residential",
    "mobile home park",
    "mountain",
    "overpass",
    "palace",
    "parking lot",
    "railway",
    "railway station",
    "rectangular farmland",
    "river",
    "roundabout",
    "runway",
    "sea ice",
    "ship",
    "snowberg",
    "sparse residential",
    "stadium",
    "storage tank",
    "tennis court",
    "terrace",
    "thermal power station",
    "wetland",
]

templates = [
    lambda c: f"satellite imagery of {c}.",
    lambda c: f"aerial imagery of {c}.",
    lambda c: f"satellite photo of {c}.",
    lambda c: f"aerial photo of {c}.",
    lambda c: f"satellite view of {c}.",
    lambda c: f"aerial view of {c}.",
    lambda c: f"satellite imagery of a {c}.",
    lambda c: f"aerial imagery of a {c}.",
    lambda c: f"satellite photo of a {c}.",
    lambda c: f"aerial photo of a {c}.",
    lambda c: f"satellite view of a {c}.",
    lambda c: f"aerial view of a {c}.",
    lambda c: f"satellite imagery of the {c}.",
    lambda c: f"aerial imagery of the {c}.",
    lambda c: f"satellite photo of the {c}.",
    lambda c: f"aerial photo of the {c}.",
    lambda c: f"satellite view of the {c}.",
    lambda c: f"aerial view of the {c}.",
]
